# Code for the two sector domain examples of the paper "Adaptive meshfree approximation for linear
# elliptic partial differential equations with PDE-greedy kernel methods" by T. Wenzel, D. Winkle, G. Santin, B. Haasdonk


# Some imports
import numpy as np
from datetime import datetime
from matplotlib import pyplot as plt

from vkoga_pde.kernels_PDE import cubicMatern_laplace
from vkoga_pde.vkoga_PDE import VKOGA_PDE
from experiments_pde_vkoga.utilities_pacman import sample_domain_pacman, get_function_pacman


np.random.seed(1)

## Create some data: Circle without cone
dim=2
N1 = int(4e5)
N2 = int(3e3)
maxIter = 1305       # 341, 1305

## Pick a kernel
kernel = cubicMatern_laplace(dim=2)


## Set up lists to loop over
list_alpha = [2/3]
list_str_function = ['singular']                # 'smooth'
list_beta = [1]
list_weight_dirichlet = [10**3]
n_iter = len(list_alpha) * len(list_str_function) * len(list_beta) * len(list_weight_dirichlet)



## Actually start the computations
idx_counter = 0 
dic_results = {}
for alpha in list_alpha:
    dic_results[alpha] = {}

    # Sample the domain
    X1, X2, X1_grid = sample_domain_pacman(N1, N2, alpha)

    for str_function in list_str_function:

        dic_results[alpha][str_function] = {}

        # Define some functions: Example from Bernard
        u, f = get_function_pacman(str_function, alpha)

        # Compute solution on fine grid
        u_X1_grid = u(X1_grid)

        # Define right hand side values
        y1, y2 = f(X1), u(X2)
        y1_sol = u(X1)

        for beta in list_beta:

            dic_results[alpha][str_function][beta] = {}

            for weight_dirichlet in list_weight_dirichlet:

                dic_results[alpha][str_function][beta][weight_dirichlet] = {}

                # Compute stuff
                model_pde_vkoga = VKOGA_PDE(kernel=kernel, beta=beta, weight_dirichlet=weight_dirichlet, verbose=True)
                _ = model_pde_vkoga.fit(X1, y1, X2, y2, maxIter=maxIter, y1_sol=y1_sol)

                # Store values
                dic_results[alpha][str_function][beta][weight_dirichlet]['f_sol'] = np.array(model_pde_vkoga.train_hist['f sol'])
                dic_results[alpha][str_function][beta][weight_dirichlet]['f'] = np.array(model_pde_vkoga.train_hist['f'])
                dic_results[alpha][str_function][beta][weight_dirichlet]['n_ctrs_bdry'] = len(model_pde_vkoga.ind_ctrs2)

                idx_counter += 1
                
                # save dictionary
                print(datetime.now().strftime("%H:%M:%S"), 'Computation {}/{} finished.'.format(idx_counter, n_iter))



# Compute some test errors (to be sure that there is no overfitting!)
y_pred_s_ = model_pde_vkoga.predict_s(X1_grid)
diff_s_ = y_pred_s_ - u(X1_grid)

y_pred_Ls_ = model_pde_vkoga.predict_Ls(X1_grid)
diff_Ls_ = y_pred_Ls_ - f(X1_grid)


print('Max error on grid: ', np.max(np.abs(diff_s_)))
print('L2 error on grid: ', np.linalg.norm(diff_s_)/np.sqrt(len(diff_s_)))



## Plot the points
plt.figure(11)
plt.clf()
plt.plot(model_pde_vkoga.ctrs_[model_pde_vkoga.ind_ctrs1][:, 0], model_pde_vkoga.ctrs_[model_pde_vkoga.ind_ctrs1][:, 1], 'ro', markersize=3)
plt.plot(model_pde_vkoga.ctrs_[model_pde_vkoga.ind_ctrs2][:, 0], model_pde_vkoga.ctrs_[model_pde_vkoga.ind_ctrs2][:, 1], 'kx', markersize=10)
# axis ratio equal
plt.gca().set_aspect('equal')
plt.show(block=False)


## Plot the decay of the approximation errors
plt.figure(12)
plt.clf()
plt.plot(model_pde_vkoga.train_hist['f sol'])
# plot horizonal line at np.max(np.abs(diff_s))
# plt.axhline(y=np.max(np.abs(diff_s_)), color='r', linestyle='-')
plt.yscale('log')
plt.xscale('log')
plt.title('|u-s_n|')
plt.show(block=False)


plt.figure(13)
plt.clf()
plt.plot(np.array(model_pde_vkoga.train_hist['f'])**(1/2))
# plot horizonal line at np.max(np.abs(diff_Ls))
# plt.axhline(y=np.max(np.abs(diff_Ls_)), color='r', linestyle='-')
plt.yscale('log')
plt.xscale('log')
plt.title('|L(u-s_n)|')
plt.show(block=False)



