# Code for the high dimensional example of the paper "Adaptive meshfree approximation for linear
# elliptic partial differential equations with PDE-greedy kernel methods" by T. Wenzel, D. Winkle, G. Santin, B. Haasdonk


# Some imports
import numpy as np
from datetime import datetime
from matplotlib import pyplot as plt

from vkoga_pde.kernels_PDE import cubicMatern_laplace, Gaussian_laplace
from vkoga_pde.vkoga_PDE import VKOGA_PDE


np.random.seed(1)


# Create data set: Square domain
dim = 12

scale_domain = 1/np.sqrt(dim)
X1 = scale_domain * np.random.rand(int(1e5), dim)
X2 = scale_domain * np.random.rand(2*dim*400, dim)

# Set boundary values
for idx_dim in range(dim):
    X2[2*idx_dim*400 : (2*idx_dim+1)*400, idx_dim] = 0
    X2[(2*idx_dim+1) * 400 : (2*idx_dim + 2) * 400, idx_dim] = scale_domain

X1_test = scale_domain * np.random.rand(int(1e5), dim)


u = lambda x: np.sum(x * x, axis=1, keepdims=True) + 1      # solution u(x) = 1/d norm(x)^2 => max in 0,1 on cube
f = lambda x: -2 * dim * np.ones((x.shape[0], 1))           # f = -2 dim on unit cube


# Define right hand side values
y1 = f(X1)
y1_sol = u(X1)
y2 = u(X2)


# Initialize and run models
kernel = cubicMatern_laplace(dim=dim)
maxIter = 1001

list_beta = [1]
list_weightings = [1e0, 1e3, 1e5]
n_iter = len(list_beta) * len(list_weightings)


## Actually start the computations
idx_counter = 0 
dic_results_train = {}
list_models = []

for beta in list_beta:

    dic_results_train[beta] = {}

    for weight_dirichlet in list_weightings:

        dic_results_train[beta][weight_dirichlet] = {}
        
        # Compute stuff
        model_pde_vkoga = VKOGA_PDE(kernel=kernel, beta=beta, weight_dirichlet=weight_dirichlet, verbose=True)
        _ = model_pde_vkoga.fit(X1, y1, X2, y2, maxIter=maxIter, y1_sol=y1_sol)

        list_models.append(model_pde_vkoga)

        # Store values
        dic_results_train[beta][weight_dirichlet]['f_sol'] = np.array(model_pde_vkoga.train_hist['f sol'])
        dic_results_train[beta][weight_dirichlet]['f'] = np.array(model_pde_vkoga.train_hist['f'])
        dic_results_train[beta][weight_dirichlet]['n_ctrs_bdry'] = len(model_pde_vkoga.ind_ctrs2)

        idx_counter += 1

        print(datetime.now().strftime("%H:%M:%S"), 'Computation {}/{} finished.'.format(idx_counter, n_iter))





## For intermediate model sizes, compute test errors
list_n_exp_size = list(np.geomspace(10, model_pde_vkoga.ctrs_.shape[0], 20, dtype=int))

dic_results_test = {}
for beta in list_beta:

    dic_results_test[beta] = {}

    for idx_model, model in enumerate(list_models):     # this corresponds to the different weightings!!

        print(datetime.now().strftime("%H:%M:%S"), 'Computation for model {}/{} started.'.format(idx_model, len(list_models)))

        dic_results_test[beta][list_weightings[idx_model]] = {}
        dic_results_test[beta][list_weightings[idx_model]]['list_s_test_pred'] = []
        dic_results_test[beta][list_weightings[idx_model]]['list_Ls_test_pred'] = []
        

        for n_expansion_size in list_n_exp_size:

            print(datetime.now().strftime("%H:%M:%S"), 'Computation n_expansion_size = {} started.'.format(n_expansion_size))

            # Restrict indices to intermediate model sizes
            ind_ctrs1 = [idx for idx in model.ind_ctrs1 if idx < n_expansion_size]
            ind_ctrs2 = [idx for idx in model.ind_ctrs2 if idx < n_expansion_size]
            ind_ctrs3 = [idx for idx in model.ind_ctrs3 if idx < n_expansion_size]

            # Compute coefficients for the restricted problem
            coef_ = model.Cut_[:n_expansion_size, :n_expansion_size].transpose() @ model.c[:n_expansion_size]

            # predict s (code taken from PDE VKOGA)
            s_test_pred = model.kernel.d2_eval(X1_test, model.ctrs_[ind_ctrs1]) @ coef_[ind_ctrs1] \
                            + model.kernel.eval(X1_test, model.ctrs_[ind_ctrs2]) @ coef_[ind_ctrs2] \
                            + model.kernel.mixed_k_n(X1_test, model.ctrs_[ind_ctrs3], model.n_ctrs_[ind_ctrs3]) @ coef_[ind_ctrs3]
            

            # predict Ls (code taken from PDE VKOGA)
            Ls_test_pred = model.kernel.dd_eval(X1_test, model.ctrs_[ind_ctrs1]) @ coef_[ind_ctrs1] \
                            + model.kernel.d1_eval(X1_test, model.ctrs_[ind_ctrs2]) @ coef_[ind_ctrs2] \
                            + model.kernel.mixed_L_n(X1_test, model.ctrs_[ind_ctrs3], model.n_ctrs_[ind_ctrs3]) @ coef_[ind_ctrs3]
            

            # Compute errors
            diff_s = s_test_pred - u(X1_test)
            diff_Ls = Ls_test_pred - f(X1_test)

            # Append errors
            dic_results_test[beta][list_weightings[idx_model]]['list_s_test_pred'].append(np.max(np.abs(diff_s)))
            dic_results_test[beta][list_weightings[idx_model]]['list_Ls_test_pred'].append(np.max(np.abs(diff_Ls)))





plt.figure(1001)
plt.clf()
for weight in list_weightings:
    plt.plot(list_n_exp_size, dic_results_test[beta][weight]['list_s_test_pred'], 'x-', label='{:.1e}'.format(weight))
plt.yscale('log')
plt.xscale('log')
plt.title('Max error on test set')
plt.legend(loc='lower left')
plt.show(block=False)



plt.figure(1002)
plt.clf()
for weight in list_weightings:
    plt.plot(list_n_exp_size, dic_results_test[beta][weight]['list_Ls_test_pred'], 'x-', label='{:.1e}'.format(weight))
plt.yscale('log')
plt.xscale('log')
plt.title('Max L-error on test set')
plt.legend(loc='lower left')
plt.legend()
plt.show(block=False)








